/* srbrl.c
 *
 * Copyright 2001, 2002 Sun Microsystems, Inc.,
 * Copyright 2001, 2002 BAUM Retec, A.G.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <netinet/in.h>
#include <netdb.h>

#include <glib.h>
#include <linc/linc.h>
#include "libsrconf.h"
#include "srbrl.h"
#include "brlinp.h"
#include <string.h>
#include <unistd.h>
#include "../braille/libbrl/brlxmlapi.h"
#include "srpres.h"


#define MIN_UDP_PORT 		1024
#define MAX_UDP_PORT 		30000


#define REMOTE_CONFIG_PATH 	"config/remote"
#define BRLMON_CONFIG_PATH 	"config/brlmon"

#define MIN_BRAILLE_PORT 1
#define MAX_BRAILLE_PORT 4


/*#define _BRLMON_DEBUG_*/

gchar *src_braille_device 	  = NULL;
gint  src_braille_port 		  = -1;
gint  src_braille_position_sensor = -1;
gint  src_braille_optical_sensor  = -1;
gint  src_braille_offset = 0;

extern gboolean src_use_braille;
extern gboolean src_use_braille_monitor;
extern void brl_input_event (BrlInEvent *event);
extern gchar *fmt_for_xml (gchar *, gchar *, gchar *);

gint 	brlmon_sd, 	/* BrlMonitor socket descriptor*/
	remote_sd;	/* RemoteBraille socket descriptor*/


extern struct sockaddr * 
linc_protocol_get_sockaddr (const LINCProtocolInfo *proto,
			    const char             *hostname,
			    const char             *service,
			    LincSockLen            *saddr_len);

extern void
linc_protocol_destroy_addr (const LINCProtocolInfo  *proto,
			    int                      fd,
			    struct sockaddr         *saddr);

/* socket structures */
#ifdef INET6
    static struct sockaddr_in6	client_addr, 
				*remote_brlmon_addr, 
				*remote_braille_addr, 
				rclient_addr;
    #define sockaddrin 	sockaddr_in6
    #define FL_AF_INET 	AF_INET6
    #define IP_TYPE	"IPv6"
#else
    static struct sockaddr_in 	client_addr, 
				*remote_brlmon_addr, 
				*remote_braille_addr, 
				rclient_addr;
    #define sockaddrin 	sockaddr_in
    #define FL_AF_INET 	AF_INET
    #define IP_TYPE	"IPv4"
#endif
static LINCProtocolInfo 	*remote_proto;
static LINCProtocolInfo 	*brlmon_proto;
GIOChannel 			*udp_giochh;

/* state flag of remote modules */
static gboolean remotesockstate = FALSE;
static gboolean brlmonsockstate = FALSE;

/* length of used ip addresses */
static LincSockLen remote_brlmon_addr_len;
static LincSockLen remote_braille_addr_len;

static gboolean 
udp_sock_glib_cb(GIOChannel *source, 
		GIOCondition condition, 
		gpointer data)
{	
    BrlInEvent event;
    BrlPackage package;
    gint clilen;
    gint n;

/*    if (!remotesockstate) return FALSE; */
    
    clilen = sizeof(rclient_addr);
    
    n = recvfrom(remote_sd, &package,sizeof(package), 0, 
		 (struct sockaddr *) &rclient_addr, &clilen);
    
    if (n < 1) return FALSE;
        
    event.event_type = package.eventtype;

    switch (event.event_type)
    {
	case BIET_KEY:	
	    (event.event_data).key_codes = g_strdup(package.source);
	    break;

	case BIET_SENSOR:
	    (event.event_data).switch_codes = g_strdup(package.source);
	    break;
	
	case BIET_SWITCH:
	    (event.event_data).sensor_codes = g_strdup(package.source);
	    break;

	default:
	    break;
    }
    
    brl_input_event(&event);
        
    return TRUE;
}

#define MAX_NO 255
#define MIN_NO 0
#define MAX_DIGIT 3

static gboolean
brl_is_ipv4(const gchar *ip)
{
    gchar number[MAX_DIGIT + 1];
    gint len;
    gint digit 	= 0;
    gint dots  	= 0;
    gint no    	= 0;
    
    if (!ip) return FALSE;
    
    len = strlen (ip);
    
    while (len != 0)
    {
	if (*ip == '.')
	{
	    number[digit] = '\0';
	    
	    no = atoi(number);
	    
	    if (no < MIN_NO || no > MAX_NO) return FALSE;
	    
	    digit = 0;
	    
	    if (++dots > MAX_DIGIT) return FALSE;
	}
	else
	if (g_ascii_isdigit(*ip))
	{
	    number[digit] = *ip;
	    
	    if (++digit > MAX_DIGIT) return FALSE;
	}
	else
	    return FALSE;
	
	len--;
	ip++;
    }
    
    if (dots != MAX_DIGIT) return FALSE;
    
    number[digit] = '\0';
    
    no = atoi (number);
    
    if (no < MIN_NO || no > MAX_NO) return FALSE;
        
    return TRUE;
}

#ifdef INET6
#define MAX_DOUBLEDOT 	7
#define MAX_HEXA_DIGIT 	4

static gboolean
brl_is_ipv6(const gchar *ip)
{
    gchar *tmp;
    gint adr  = 0;
    gint len;
    gint ddot = 0;
    gboolean compres = FALSE;
    
    if (!ip) return FALSE;
    
    len = strlen (ip);
    
    while (len != 0)
    {
	if (*ip == ':')
	{
	    if (ip - tmp == 1)
	    {
		if (compres) return FALSE;
		compres = TRUE;
	    }
	    tmp = (gchar*)ip;
	    adr = 0;
	    if (++ddot > MAX_DOUBLEDOT) return FALSE;
	}
	else
	if (*ip == '.') 
	{
	    if (ddot >= MAX_DOUBLEDOT)  return FALSE;
	    return brl_is_ipv4 (tmp + 1);
	}
	else
	if (g_ascii_isxdigit(*ip))
	{
	    if (++adr > MAX_HEXA_DIGIT) return FALSE;
	}
	else
	    return FALSE;
	    
	len--;
	ip++;
    }
    
    return TRUE;
}
#endif

static gboolean
brl_valid_ip_address(const gchar *ip)
{
#ifdef INET6
    return brl_is_ipv6 (ip);
#else
    return brl_is_ipv4 (ip);
#endif
}

static gboolean
brl_remote_udp_init()
{
    gint defport 	= DEFAULT_REMOTE_PORT;
    gint brlport;
    gchar *brlip 	= NULL;

    if (!srconf_get_data_with_default ("port",CFGT_INT,(gpointer)&brlport,(gpointer)&defport,REMOTE_CONFIG_PATH)) {
	    g_warning ("Can not get UDP Port number for Remote from ~/.gconf/apps/gnopernicus/config/remote/gconf.xml config file");
	    remotesockstate = FALSE; 
	    return FALSE; 	    
	}

    if (!srconf_get_data_with_default ("ip",CFGT_STRING,(gpointer)&brlip,(gpointer)DEFAULT_REMOTE_IP,REMOTE_CONFIG_PATH)) {
	    g_warning ("Can not get IP addres for Remote from ~/.gconf/apps/gnopernicus/config/remote/gconf.xml config file");
	    remotesockstate = FALSE; 
	    return FALSE; 
	}
	    
    if (brlport	< MIN_UDP_PORT || brlport > MAX_UDP_PORT) {
	    g_warning ("Invalid UDP port number:%d",brlport);
	    brlport = DEFAULT_REMOTE_PORT;
	}


    if (!brl_valid_ip_address (brlip)) {
	    g_warning ("Invalid IP (%s) address:%s",IP_TYPE,brlip);
	    fprintf (stderr,"Use %s localhost address %s\n",IP_TYPE,DEFAULT_REMOTE_IP);
	    free (brlip);
	    brlip = g_strdup (DEFAULT_REMOTE_IP);
	}
	
    remote_proto = linc_protocol_find (IP_TYPE);
    
    if (remote_proto == NULL || remote_proto->get_sockinfo == NULL) {
	    g_warning ("Invalid protocol type");
	    remotesockstate = FALSE; 
	    free (brlip);
	    return FALSE;
	}
	
    remote_braille_addr = (struct sockaddrin *) linc_protocol_get_sockaddr (
		remote_proto, 
		brlip,
		g_strdup_printf ("%d",brlport), 
		(LincSockLen*)&remote_braille_addr_len);    


    
    if ((remote_sd = socket (FL_AF_INET, SOCK_DGRAM, 0)) < 0) {
	    g_warning ("Can not open socket");
	    
	    remotesockstate = FALSE;
	    
	    g_free (brlip);	    
	    brlip = NULL;
	    
	    return FALSE;
	}
 
    bzero ((gchar*) &rclient_addr, sizeof (struct sockaddrin));
#ifdef INET6
    rclient_addr.sin6_family 	= FL_AF_INET;
    rclient_addr.sin6_addr	= in6addr_any;
    rclient_addr.sin6_port 	= htons (0);
#else
    rclient_addr.sin_family 	= FL_AF_INET;
    rclient_addr.sin_addr.s_addr= htonl (INADDR_ANY);
    rclient_addr.sin_port 	= htons (0);
#endif

    if (bind (remote_sd, (struct sockaddr *) &rclient_addr, sizeof (rclient_addr)) < 0) {
	    g_warning ("Can not bind Remote host.");
	    
	    remotesockstate = FALSE;
	    
	    g_free (brlip);
	    brlip = NULL;
	    
	    return FALSE;
	}
    
    remotesockstate = TRUE;

    fprintf (stderr,"\n** Remote Braille Device initialization succeded for %s host"
			" on port %d.", brlip, brlport);
    fprintf (stderr, "\n   To see something you must have remote server started.");
    fprintf (stderr, "\n   To change one or more settings use gconf-editor tool."
			    "\n     Changes must be maded in .../apps/gnopernicus/config/Remote.\n");

    g_free(brlip);
    brlip = NULL;

    udp_giochh = g_io_channel_unix_new (remote_sd);
    
    g_io_add_watch (udp_giochh, G_IO_IN | G_IO_PRI, udp_sock_glib_cb, NULL);	
    
    return TRUE;
}

static gboolean 
brl_mon_udp_init()
{
    gint monport;
    gint defport = DEFAULT_BRLMON_PORT;
    gchar *monip = NULL;

    if (!srconf_get_data_with_default ("port",CFGT_INT,(gpointer)&monport,(gpointer)&defport,BRLMON_CONFIG_PATH)){
	    g_warning("Can not get UDP Port number for Braille monitor from ~/.gconf/apps/gnopernicus/config/brlmon/gconf.xml config file%d",monport);
	    brlmonsockstate = FALSE; 
	    return FALSE;
	}
    
    if (!srconf_get_data_with_default ("ip",CFGT_STRING,(gpointer)&monip,(gpointer)DEFAULT_BRLMON_IP,BRLMON_CONFIG_PATH)) {
	    g_warning ("Can not get IP addres for Braille monitor from ~/.gconf/apps/gnopernicus/config/brlmon/gconf.xml config file");
	    brlmonsockstate = FALSE; 
	    return FALSE;
	}
	    
    if (monport	< MIN_UDP_PORT || monport > MAX_UDP_PORT) {
	    g_warning ("Invalid UDP port number:%d",monport);
	    monport = DEFAULT_BRLMON_PORT;
	}

    if (!brl_valid_ip_address (monip)) {
	    g_warning ("Invalid IP (%s) address:%s",IP_TYPE,monip);
	    fprintf (stderr,"Use %s localhost address %s\n",IP_TYPE,DEFAULT_BRLMON_IP);
    	    free (monip);
	    monip = g_strdup (DEFAULT_BRLMON_IP);
	}
	
    brlmon_proto = linc_protocol_find (IP_TYPE);
    
    if (brlmon_proto == NULL || brlmon_proto->get_sockinfo == NULL) {
	    g_warning ("Invalid protocol type");
	    brlmonsockstate = FALSE; 
	    free (monip);
	    return FALSE;
	}

    remote_brlmon_addr = (struct sockaddrin *) linc_protocol_get_sockaddr (
		brlmon_proto, 
		monip,
		g_strdup_printf ("%d",monport), 
		(LincSockLen*)&remote_brlmon_addr_len);		        

    if ((brlmon_sd = socket (FL_AF_INET, SOCK_DGRAM, 0)) < 0) {
	    g_warning ("Can not open socket");
	    
	    brlmonsockstate = FALSE;
	    
	    g_free (monip);
	    monip = NULL;
	    
	    return FALSE;
	}
	    
    bzero ((gchar*) &client_addr, sizeof (struct sockaddrin));
#ifdef INET6
    client_addr.sin6_family 	= FL_AF_INET;
    client_addr.sin6_addr 	= in6addr_any;
    client_addr.sin6_port 	= htons (0);
#else
    client_addr.sin_family 	= FL_AF_INET;
    client_addr.sin_addr.s_addr = htonl (INADDR_ANY);
    client_addr.sin_port 	= htons (0);
#endif

    if (bind (brlmon_sd, (struct sockaddr *) &client_addr, sizeof (client_addr)) < 0) { 
	    g_warning ("Can not bind Braille Monitor host.");
	    
	    brlmonsockstate = FALSE;
	    
	    g_free (monip);
	    monip = NULL;
	    
	    return FALSE;
	}
    
    fprintf (stderr,"\n** Braille Monitor initialization succeded for %s host"
			" on port %d.", monip,monport);
    fprintf (stderr, "\n   To see something you must have Braille Monitor"
			" server started.");
    fprintf (stderr, "\n   To change one or more settings use gconf-editor tool."
			    "\n     Changes must be maded in .../apps/gnopernicus/config/BrlMon.\n");
    
    g_free(monip);
    
    monip = NULL;
    
    brlmonsockstate = TRUE;
    
    return TRUE;
}

void 
src_brlmon_terminate()
{
    if (brlmonsockstate == TRUE) 
    {
	src_brlmon_quit ();
	close (brlmon_sd);
/*#ifdef INET6
	linc_protocol_destroy_addr (brlmon_proto,
				    brlmon_sd,
				    &(remote_brlmon_addr->sin6_addr));
#else
	linc_protocol_destroy_addr (brlmon_proto,
				    brlmon_sd,
				    ((remote_brlmon_addr->sin_addr).s_addr));
#endif*/
	brlmonsockstate = FALSE;
	}
}

static void 
brl_remote_udp_terminate()
{
    if (remotesockstate == TRUE) {
	close (remote_sd);
/*#ifdef INET6
	linc_protocol_destroy_addr (remote_proto,
				    remote_sd,
				    &(remote_braille_addr->sin6_addr));
#else
	linc_protocol_destroy_addr (remote_proto,
				    remote_sd,
				    &((remote_braille_addr->sin_addr).s_addr));
#endif*/
	remotesockstate = FALSE;
	}
}


static void 
brl_send_package(gchar *sendtmp,gint len)
{
    gint rc = 0;
    if (brlmonsockstate)
    {
	rc = sendto (brlmon_sd, sendtmp, len, 0,
		    (struct sockaddr *) remote_brlmon_addr,
		    remote_brlmon_addr_len);
	if ( rc == -1)	g_warning ("Can not send data to Braille Monitor");
    }
    if (remotesockstate)
    {
	rc = sendto (remote_sd, sendtmp, len, 0,
		    (struct sockaddr *) remote_braille_addr,
		    remote_braille_addr_len);
	if ( rc == -1)	g_warning ("Can not send data to Remote Braille");
    }
}

static void 
src_brl_chop_output(gchar *buff)
{
    gint len 		= strlen (buff);
    gint no_of_pack 	= len / (BUFFER_SIZE - 1);
    gint rest 		= len % (BUFFER_SIZE - 1);
    gchar *tmp 		= buff;
    gchar sendtmp[BUFFER_SIZE];
    gint iter;
    
    if (rest != 0) no_of_pack++;
        
    memset(sendtmp, 0x0, sizeof (sendtmp));
    
    sendtmp[0] = 0x02;
    
    brl_send_package(sendtmp, sizeof (sendtmp));
    
    for(iter = 0; iter < no_of_pack; ++iter)
    {
	memset (sendtmp, 0x0, sizeof (sendtmp));
		
	strncpy (sendtmp, tmp, BUFFER_SIZE - 1);
			
	tmp = tmp + (BUFFER_SIZE - 1);
		
	brl_send_package (sendtmp, sizeof (sendtmp));
    }
	
    memset (sendtmp, 0x0, sizeof (sendtmp));
    
    sendtmp[0] = 0x03;
    
    brl_send_package (sendtmp, sizeof (sendtmp));
}


void 
src_brlmon_quit()
{
    src_brlmon_send("<QUIT>");
}

void 
src_brlmon_send (gchar *monoutput)
{
    if (monoutput)
    {
	src_brl_chop_output (monoutput);
    }
}

void 
src_brlmon_show (gchar *message)
{
    gchar *brlout = NULL;
    gchar *text 	= NULL;
    gchar offset[5];
	
    text = src_xml_format ("TEXT", NULL, message);
    sprintf (offset, "%d", src_braille_offset);
    brlout = g_strconcat (
		"<BRLOUT language=\"us\" bstyle=\"8\" clear=\"true\">\n",
		/* main display */
		"<BRLDISP role=\"main\" offset=\"", offset, "\" clear=\"true\">\n",
		text,
		"</BRLDISP>\n",
		/* nothing to do on state display here*/
		"</BRLOUT>\n", 
		NULL);
	
    src_brlmon_send  (brlout);
    g_free (brlout);						
    g_free(text);
}



void 
src_braille_send (gchar *brloutput)
{
    if (brloutput)
    {
	brl_xml_output(brloutput, strlen(brloutput));
/*	src_brlmon_send(brloutput);*/
	/* fprintf (stderr, "\n%s", brloutput); */
    }
}

void 
src_braille_show (gchar *message)
{
    gchar *brlout = NULL;
    gchar *text 	= NULL;
    gchar offset[5];
	
    text = src_xml_format ("TEXT", NULL, message);
    sprintf (offset, "%d", src_braille_offset);	
    brlout = g_strconcat (
		"<BRLOUT language=\"us\" bstyle=\"8\" clear=\"true\">\n",
		/* main display */
		"<BRLDISP role=\"main\" offset=\"", offset, "\" clear=\"true\">\n",
		text,
		"</BRLDISP>\n",
		/* nothing to do on state display here*/
		"</BRLOUT>\n", 
		NULL);
	
    src_braille_send (brlout);
    g_free (brlout);						
    g_free(text);
}

gboolean
src_braille_restart ()
{
    if (src_use_braille)
    {
	gchar *device = g_strdup (src_braille_device);
	gint port = src_braille_port;
	/* terminate Braille */
	src_braille_terminate ();
    	/* and then start it again */
	src_braille_device = device;
	src_braille_port = port;	
	src_use_braille = src_braille_init ();
    }
    if (src_use_braille_monitor)
    {
    }
    return src_use_braille;
}

extern void brl_xml_input_proc (gchar *, gint);

gboolean
src_braille_monitor_init ()
{
    brl_mon_udp_init ();
    brl_remote_udp_init();
    
    if (!remotesockstate && !brlmonsockstate)
	{
	    fprintf (stderr, "**Can't initialize any module.\n");
	    srconf_unset_key (SRCORE_BRAILLE_MONITOR_SENSITIVE, SRCORE_PATH);
	    SET_SRCORE_CONFIG_DATA (SRCORE_BRAILLE_MONITOR_SENSITIVE, CFGT_BOOL, &brlmonsockstate);
	}
    return TRUE;
}

gboolean 
src_braille_init ()
{
    gboolean brlval = FALSE;
    
    brl_in_xml_init (brl_input_event);
    if (!src_braille_device)
    {
	gchar *default_device = DEFAULT_BRAILLE_DEVICE;
	GET_BRAILLE_CONFIG_DATA_WITH_DEFAULT (BRAILLE_DEVICE, CFGT_STRING, 
		&src_braille_device, default_device);
    }
	
    if (MIN_BRAILLE_PORT > src_braille_port || MAX_BRAILLE_PORT < src_braille_port) /* just for sure */
    {
	gint default_port = DEFAULT_BRAILLE_PORT_NO;
	GET_BRAILLE_CONFIG_DATA_WITH_DEFAULT (BRAILLE_PORT_NO, CFGT_INT, 
		    &src_braille_port, &default_port);
	if (MIN_BRAILLE_PORT > src_braille_port || MAX_BRAILLE_PORT < src_braille_port) 
	{
	    src_braille_port = DEFAULT_BRAILLE_PORT_NO;
	    SET_BRAILLE_CONFIG_DATA (BRAILLE_PORT_NO, CFGT_INT, &src_braille_port);
	}
    }

    if (!brl_xml_init(src_braille_device, src_braille_port, brl_xml_input_proc)) /* take Braille display from Gnopi... */
    {
	brlval = FALSE;
	fprintf(stderr, "\n** BRAILLE initialization failed for %s on port %d",
					src_braille_device, src_braille_port);
	fprintf (stderr, "\n   To change one or more settings use gnopernicus UI\n");
    }
    else
    {
	brlval = TRUE;
	fprintf (stderr, "\n** BRAILLE initialization succeded for %s device"
				    " on port %d.", src_braille_device, src_braille_port);
	fprintf (stderr, "\n   To see something you must have a Braille device"
				    " connected to your computer.");
	fprintf (stderr, "\n   To change one or more settings use gnopernicus UI.\n");
	
    }
    srconf_unset_key (SRCORE_BRAILLE_SENSITIVE, SRCORE_PATH);
    SET_SRCORE_CONFIG_DATA (SRCORE_BRAILLE_SENSITIVE, CFGT_BOOL, &brlval);
    return brlval;
}

void
src_braille_monitor_terminate ()
{
    src_brlmon_terminate ();
    brl_remote_udp_terminate();
}

void
src_braille_terminate ()
{
    brl_xml_terminate ();
    brl_in_xml_terminate ();
	
    if (src_braille_device)
	g_free (src_braille_device);
    src_braille_device = NULL;
    src_braille_port = -1;
}
